"""
This is your main interface to the Authorize.net API, where you feed in your
credentials and then interact to create transactions and so forth. You will
need to sign up for your own developer account and credentials at
Authorize.net_.

.. warning::

    To use the saved card and recurring billing functionality, you must have
    either or both set up in your Authorize.net account. You must sign up your
    account for the CIM (Customer Information Manager) service and/or the ARB
    (Automated Recurring Billing) service, each of which may be an additional
    monthly charge. See the :ref:`authorize-net-documentation` for additional
    information.

.. _Authorize.net: http://developer.authorize.net/

"""
from uuid import uuid4

from authorize.apis.customer import CustomerAPI
from authorize.apis.recurring import RecurringAPI
from authorize.apis.transaction import TransactionAPI


class AuthorizeClient(object):
    """
    Instantiate the client with your login ID and transaction key from
    Authorize.net.

    The ``debug`` option determines whether to use debug mode
    for the APIs. This should be ``True`` in development and staging, and
    should be ``False`` in production when you want to actually process credit
    cards. You will need to pass in the appropriate login credentials
    depending on debug mode. The ``test`` option determines whether to run
    the standard API in test mode, which should generally be left ``False``,
    even in development and staging environments.
    """
    def __init__(self, login_id, transaction_key, debug=True, test=False):
        self.login_id = login_id
        self.transaction_key = transaction_key
        self.debug = debug
        self.test = test
        self._transaction = TransactionAPI(login_id, transaction_key,
            debug, test)
        self._recurring = RecurringAPI(login_id, transaction_key, debug, test)
        self._customer = CustomerAPI(login_id, transaction_key, debug, test)

    def card(self, credit_card, address=None, email=None):
        """
        To work with a credit card, pass in a
        :class:`CreditCard <authorize.data.CreditCard>` instance, and
        optionally an :class:`Address <authorize.data.Address>` instance. This
        will return an
        :class:`AuthorizeCreditCard <authorize.client.AuthorizeCreditCard>`
        instance you can then use to execute transactions.
        ``email`` is only required for those using European payment processors.
        """
        return AuthorizeCreditCard(self, credit_card, address=address,
                                   email=email)

    def transaction(self, uid):
        """
        To perform an action on a previous transaction, pass in the ``uid`` of
        that transaction as a string. This will return an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance you can then use to settle, credit or void that transaction.
        """
        return AuthorizeTransaction(self, uid)

    def saved_card(self, uid):
        """
        To create a new transaction from a saved card, pass in the ``uid`` of
        the saved card as a string. This will return an
        :class:`AuthorizeSavedCard <authorize.client.AuthorizeSavedCard>`
        instance you can then use to auth, capture, or create a credit.
        """
        return AuthorizeSavedCard(self, uid)

    def recurring(self, uid):
        """
        To update or cancel an existing recurring payment, pass in the ``uid``
        of the recurring payment as a string. This will return an
        :class:`AuthorizeRecurring <authorize.client.AuthorizeRecurring>`
        instance you can then use to udpate or cancel the payments.
        """
        return AuthorizeRecurring(self, uid)


class AuthorizeCreditCard(object):
    """
    This is the interface for working with a credit card. You use this to
    authorize or charge a credit card, as well as saving the credit card and
    creating recurring payments.

    Any operation performed on this instance returns another instance you can
    work with, such as a transaction, saved card, or recurring payment.
    """
    def __init__(self, client, credit_card, address=None, email=None):
        self._client = client
        self.credit_card = credit_card
        self.address = address
        self.email = email

    def __repr__(self):
        return '<AuthorizeCreditCard {0.credit_card.card_type} ' \
            '{0.credit_card.safe_number}>'.format(self)

    def auth(self, amount):
        """
        Authorize a transaction against this card for the specified amount.
        This verifies the amount is available on the card and reserves it.
        Returns an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance representing the transaction.
        """
        response = self._client._transaction.auth(
            amount, self.credit_card, self.address, self.email)
        transaction = self._client.transaction(response['transaction_id'])
        transaction.full_response = response
        return transaction

    def capture(self, amount):
        """
        Capture a transaction immediately on this card for the specified
        amount. Returns an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance representing the transaction.
        """
        response = self._client._transaction.capture(
            amount, self.credit_card, self.address, self.email)
        transaction = self._client.transaction(response['transaction_id'])
        transaction.full_response = response
        return transaction

    def save(self):
        """
        Saves the credit card on Authorize.net's servers so you can create
        transactions at a later date. Returns an
        :class:`AuthorizeSavedCard <authorize.client.AuthorizeSavedCard>`
        instance that you can save or use.
        """
        unique_id = uuid4().hex[:20]
        payment = self._client._customer.create_saved_payment(
            self.credit_card, address=self.address)
        profile_id, payment_ids = self._client._customer \
            .create_saved_profile(unique_id, [payment], email=self.email)
        uid = '{0}|{1}'.format(profile_id, payment_ids[0])
        return self._client.saved_card(uid)

    def recurring(self, amount, start, days=None, months=None,
            occurrences=None, trial_amount=None, trial_occurrences=None):
        """
        Creates a recurring payment with this credit card. Pass in the
        following arguments to set it up:

        ``amount``
            The amount to charge at each interval.

        ``start``
            The ``date`` or ``datetime`` at which to begin the recurring
            charges.

        ``days``
            The number of days in the billing cycle. You must provide either
            the ``days`` argument or the ``months`` argument.

        ``months``
            The number of months in the billing cycle. You must provide either
            the ``days`` argument or the ``months`` argument.

        ``occurrences`` *(optional)*
            The number of times the card should be billed before stopping. If
            not specified, it will continue indefinitely.

        ``trial_amount`` *(optional)*
            If you want to charge a lower amount for an introductory period,
            specify the amount.

        ``trial_occurrences`` *(optional)*
            If you want to charge a lower amount for an introductory period,
            specify the number of occurrences that period should last.

        Returns an
        :class:`AuthorizeRecurring <authorize.client.AuthorizeRecurring>`
        instance that you can save, update or delete.
        """
        uid = self._client._recurring.create_subscription(
            self.credit_card, amount, start, days=days, months=months,
            occurrences=occurrences, trial_amount=trial_amount,
            trial_occurrences=trial_occurrences)
        return self._client.recurring(uid)


class AuthorizeTransaction(object):
    """
    This is the interface for working with a previous transaction. It is
    returned by many other operations, or you can save the transaction's
    ``uid`` and reinstantiate it later.

    You can then use this transaction to settle a previous authorization,
    credit back a previous transaction, or void a previous authorization. Any
    such operation returns another transaction instance you can work with.

    Additionally, if you need to access the full raw result of the transaction
    it is stored in the ``full_response`` attribute on the class.
    """
    def __init__(self, client, uid):
        self._client = client
        self.uid = uid

    def __repr__(self):
        return '<AuthorizeTransaction {0.uid}>'.format(self)

    def settle(self, amount=None):
        """
        Settles this transaction if it is a previous authorization. If no
        ``amount`` is specified, the full amount will be settled; if a lower
        ``amount`` is provided, the lower amount will be settled; if a higher
        ``amount`` is given, it will result in an error. Returns an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance representing the settlement transaction.
        """
        response = self._client._transaction.settle(self.uid, amount=amount)
        transaction = self._client.transaction(response['transaction_id'])
        transaction.full_response = response
        return transaction

    def credit(self, card_number, amount):
        """
        Creates a credit (refund) back on the original transaction. The
        ``card_number`` should be the last four digits of the credit card
        and the ``amount`` is the amount to credit the card. Returns an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance representing the credit transaction.

        Credit transactions are bound by a number of restrictions:

        * The original transaction must be an existing, settled charge. (Note
          that this is different than merely calling the
          :meth:`AuthorizeTransaction.settle <authorize.client.AuthorizeTransaction.settle>`
          method, which submits a payment for settlement. In production,
          Authorize.net actually settles charges once daily. Until a charge is
          settled, you should use
          :meth:`AuthorizeTransaction.void <authorize.client.AuthorizeTransaction.void>`
          instead.)
        * The amount of the credit (as well as the sum of all credits against
          this original transaction) must be less than or equal to the
          original amount charged.
        * The credit transaction must be submitted within 120 days of the date
          the original transaction was settled.
        """
        response = self._client._transaction.credit(
            card_number, self.uid, amount)
        transaction = self._client.transaction(response['transaction_id'])
        transaction.full_response = response
        return transaction

    def void(self):
        """
        Voids a previous authorization that has not yet been settled. Returns
        an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance representing the void transaction.
        """
        response = self._client._transaction.void(self.uid)
        transaction = self._client.transaction(response['transaction_id'])
        transaction.full_response = response
        return transaction


class AuthorizeSavedCard(object):
    """
    This is the interface for working with a saved credit card. It is returned
    by the
    :meth:`AuthorizeCreditCard.save <authorize.client.AuthorizeCreditCard.save>`
    method, or you can save a saved card's ``uid`` and reinstantiate it later.

    You can then use this saved card to create new authorizations, captures,
    and credits. Or you can delete this card from the Authorize.net database.
    The first three operations will all return a transaction instance to work
    with.

    You can also retrieve payment information with the
    :meth:`AuthorizeSavedCard.get_payment_info <authorize.client.AuthorizeSavedCard.get_payment_info`
    method.

    You can update this information by setting it and running the
    :meth:`AuthorizeSavedCard.update <authorize.client.AuthorizeSavedCard.update>`
    method.
    """

    def __init__(self, client, uid):
        self._client = client
        self.uid = uid
        self._profile_id, self._payment_id = uid.split('|')

    def __repr__(self):
        return '<AuthorizeSavedCard {0.uid}>'.format(self)

    def auth(self, amount, cvv=None):
        """
        Authorize a transaction against this card for the specified amount.
        This verifies the amount is available on the card and reserves it.
        Returns an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance representing the transaction.
        """
        response = self._client._customer.auth(
            self._profile_id, self._payment_id, amount, cvv)
        transaction = self._client.transaction(response['transaction_id'])
        transaction.full_response = response
        return transaction

    def capture(self, amount, cvv=None):
        """
        Capture a transaction immediately on this card for the specified
        amount. Returns an
        :class:`AuthorizeTransaction <authorize.client.AuthorizeTransaction>`
        instance representing the transaction.
        """
        response = self._client._customer.capture(
            self._profile_id, self._payment_id, amount, cvv)
        transaction = self._client.transaction(response['transaction_id'])
        transaction.full_response = response
        return transaction

    def update(self, **kwargs):
        """
        Updates information about a saved card. You can use this to change the
        address associated with the card, or the name, or the user's email
        address. You may even update the expiration date.

        ``number`` *(optional)*
            An updated number for the card. In most cases, you should create a
            new card instead of using this.

        ``first_name`` *(optional)*
            The first name on the card.

        ``last_name`` *(optional)*
            The last name on the card

        ``address`` *(optional)*
            An :class:`Address <authorize.data.Address>` object that holds new
            billing address information for the card.

        ``email`` *(optional)*
            Updates the email associated with the card. Note that emails are
            actually stored on the customer profile. If you have multiple cards
            (payments) with the same profile, this will update the email for
            all of them.

        ``exp_month`` *(conditional)*
            If this option is specified, ``exp_year`` must also be specified,
            or it will be ignored.

            An integer representing the month of the card's expiration date.

        ``exp_year`` *(conditional)*
            If this option is specified, ``exp_month`` must also be specified,
            or it will be ignored.

            An integer representing the year of the card's expiration date.
        """
        settings = {'exp_month': None, 'exp_year': None}
        old_settings = self.get_payment_info()
        settings.update(old_settings)
        settings.update(**kwargs)
        self._client._customer.update_saved_payment(
            self._profile_id, self._payment_id, **settings)

    def get_payment_info(self):
        """
        Retrieves information about a card. It will return a dictionary
        containing the ``first_name`` and ``last_name`` on the card, as well
        as an ``address`` (a :class:`Address <authorize.data.Address>`
        instance), and the user's ``email``. These can be used, for instance,
        to populate a form which the user can fill to update their payment
        information.

        Note that the expiration date fields are masked by Authorize.net, and
        so are not returned by this function. Neither is the card number
        itself.
        """
        return self._client._customer.retrieve_saved_payment(
            self._profile_id, self._payment_id)

    def delete(self):
        """
        Removes this saved card from the Authorize.net database.
        """
        self._client._customer.delete_saved_payment(
            self._profile_id, self._payment_id)


class AuthorizeRecurring(object):
    """
    This is the interface for working with a recurring charge. It is returned
    by the
    :meth:`AuthorizeCreditCard.recurring <authorize.client.AuthorizeCreditCard.recurring>`
    method, or you can save a recurring payment's ``uid`` and reinstantiate it
    later.

    The recurring payment will continue charging automatically, but if you
    want to make changes to an existing recurring payment or to cancel a
    recurring payment, this provides the interface.
    """
    def __init__(self, client, uid):
        self._client = client
        self.uid = uid

    def __repr__(self):
        return '<AuthorizeRecurring {0.uid}>'.format(self)

    def update(self, amount=None, start=None, occurrences=None,
            trial_amount=None, trial_occurrences=None):
        """
        Updates the amount or status of the recurring payment. You may provide
        any or all fields and they will be updated appropriately, so long as
        none conflict. Fields work as described under the

        ``amount`` *(optional)*
            The amount to charge at each interval. Will only be applied to
            future charges.

        ``start`` *(optional)*
            The ``date`` or ``datetime`` at which to begin the recurring
            charges. You may only specify this option if the recurring charge
            has not yet begun.

        ``occurrences`` *(optional)*
            The number of times the card should be billed before stopping. If
            not specified, it will continue indefinitely.

        ``trial_amount`` *(optional)*
            If you want to charge a lower amount for an introductory period,
            specify the amount. You may specify this option only if there have
            not yet been any non-trial payments.

        ``trial_occurrences`` *(optional)*
            If you want to charge a lower amount for an introductory period,
            specify the number of occurrences that period should last. You may
            specify this option only if there have not yet been any non-trial
            payments.
        """
        self._client._recurring.update_subscription(self.uid,
            amount=amount, start=start, occurrences=occurrences,
            trial_amount=trial_amount, trial_occurrences=trial_occurrences)

    def delete(self):
        """
        Cancels any future charges from this recurring payment.
        """
        self._client._recurring.delete_subscription(self.uid)



"""
This module provides the data structures for describing credit cards and
addresses for use in executing charges.
"""

import calendar
from datetime import datetime
import re

from authorize.exceptions import AuthorizeInvalidError


CARD_TYPES = {
    'visa': r'4\d{12}(\d{3})?$',
    'amex': r'37\d{13}$',
    'mc': r'5[1-5]\d{14}$',
    'discover': r'6011\d{12}',
    'diners': r'(30[0-5]\d{11}|(36|38)\d{12})$'
}


class CreditCard(object):
    """
    Represents a credit card that can be charged.

    Pass in the credit card number, expiration date, CVV code, and optionally
    a first name and last name. The card will be validated upon instatiation
    and will raise an
    :class:`AuthorizeInvalidError <authorize.exceptions.AuthorizeInvalidError>`
    for invalid credit card numbers, past expiration dates, etc.
    """
    def __init__(self, card_number=None, exp_year=None, exp_month=None,
            cvv=None, first_name=None, last_name=None):
        self.card_number = re.sub(r'\D', '', str(card_number))
        self.exp_year = str(exp_year)
        self.exp_month = str(exp_month)
        self.cvv = str(cvv)
        self.first_name = first_name
        self.last_name = last_name
        self.validate()

    def __repr__(self):
        return '<CreditCard {0.card_type} {0.safe_number}>'.format(self)

    def validate(self):
        """
        Validates the credit card data and raises an
        :class:`AuthorizeInvalidError <authorize.exceptions.AuthorizeInvalidError>`
        if anything doesn't check out. You shouldn't have to call this
        yourself.
        """
        try:
            num = [int(n) for n in self.card_number]
        except ValueError:
            raise AuthorizeInvalidError('Credit card number is not valid.')
        if sum(num[::-2] + [sum(divmod(d * 2, 10)) for d in num[-2::-2]]) % 10:
            raise AuthorizeInvalidError('Credit card number is not valid.')
        if datetime.now() > self.expiration:
            raise AuthorizeInvalidError('Credit card is expired.')
        if not re.match(r'^\d{3,4}$', self.cvv):
            raise AuthorizeInvalidError('Credit card CVV is invalid format.')
        if not self.card_type:
            raise AuthorizeInvalidError('Credit card number is not valid.')

    @staticmethod
    def exp_time(exp_month, exp_year):
        exp_year, exp_month = int(exp_year), int(exp_month)
        return datetime(exp_year, exp_month,
            calendar.monthrange(exp_year, exp_month)[1],
            23, 59, 59)

    @property
    def expiration(self):
        """
        The credit card expiration date as a ``datetime`` object.
        """
        return self.exp_time(self.exp_month, self.exp_year)

    @property
    def safe_number(self):
        """
        The credit card number with all but the last four digits masked. This
        is useful for storing a representation of the card without keeping
        sensitive data.
        """
        mask = '*' * (len(self.card_number) - 4)
        return '{0}{1}'.format(mask, self.card_number[-4:])

    @property
    def card_type(self):
        """
        The credit card issuer, such as Visa or American Express, which is
        determined from the credit card number. Recognizes Visa, American
        Express, MasterCard, Discover, and Diners Club.
        """
        for card_type, card_type_re in CARD_TYPES.items():
            if re.match(card_type_re, self.card_number):
                return card_type


class Address(object):
    """
    Represents a billing address for a charge. Pass in the street, city, state
    and zip code, and optionally country for the address.
    """
    def __init__(self, street=None, city=None, state=None, zip_code=None,
            country='US'):
        self.street = street
        self.city = city
        self.state = state
        self.zip_code = zip_code
        self.country = country

    def __repr__(self):
        return '<Address {0.street}, {0.city}, {0.state} {0.zip_code}>' \
            .format(self)



"""
Authorize.net payment gateway interface.

For original Authorize.net API docs:

AIM (basic transactions)
    http://www.authorize.net/support/AIM_guide.pdf

CIM (saved payment)
    http://www.authorize.net/support/CIM_SOAP_guide.pdf

ARB (recurring payments)
    http://www.authorize.net/support/ARB_SOAP_guide.pdf

The client init arguments are as follows:

login_id
    The login ID provided by Authorize.net for your account.

transaction_key
    The transaction key provided by Authorize.net for your account.

debug
    Set to True to run all API calls against the Authorize.net staging
    environment. You will receive normal responses but none of your
    transactions will actually be processed. Your login_id and transaction_key
    must belong to a developer account. Use this in development.

test
    Set to True to run all API calls in test mode against the Authorize.net
    production environment. This will use your real production login_id and
    transaction_key credentials and the real production API, but it will not
    actually send transactions for processing. Use this just before deploying
    to ensure your connection and credentials are valid. (Only affects the
    basic transactions API, not saved payments or subscriptions.)

"""

from authorize.client import AuthorizeClient
from authorize.data import Address, CreditCard
from authorize.exceptions import AuthorizeConnectionError, AuthorizeError, \
    AuthorizeInvalidError, AuthorizeResponseError



class AuthorizeError(Exception):
    """Base class for all errors."""


class AuthorizeConnectionError(AuthorizeError):
    """Error communicating with the Authorize.net API."""


class AuthorizeResponseError(AuthorizeError):
    """Error response code returned from API."""


class AuthorizeInvalidError(AuthorizeError):
    """Invalid information provided."""



from datetime import date
from decimal import Decimal
from ssl import SSLError

from suds import WebFault
from suds.client import Client

from authorize.exceptions import AuthorizeConnectionError, \
    AuthorizeInvalidError, AuthorizeResponseError


PROD_URL = 'https://api.authorize.net/soap/v1/Service.asmx?WSDL'
TEST_URL = 'https://apitest.authorize.net/soap/v1/Service.asmx?WSDL'


class RecurringAPI(object):
    def __init__(self, login_id, transaction_key, debug=True, test=False):
        self.url = TEST_URL if debug else PROD_URL
        self.login_id = login_id
        self.transaction_key = transaction_key

    @property
    def client(self):
        # Lazy instantiation of SOAP client, which hits the WSDL url
        if not hasattr(self, '_client'):
            self._client = Client(self.url)
        return self._client

    @property
    def client_auth(self):
        if not hasattr(self, '_client_auth'):
            self._client_auth = self.client.factory.create(
                'MerchantAuthenticationType')
            self._client_auth.name = self.login_id
            self._client_auth.transactionKey = self.transaction_key
        return self._client_auth

    def _make_call(self, service, *args):
        # Provides standard API call error handling
        method = getattr(self.client.service, service)
        try:
            response = method(self.client_auth, *args)
        except (WebFault, SSLError) as e:
            raise AuthorizeConnectionError(e)
        if response.resultCode != 'Ok':
            error = response.messages[0][0]
            raise AuthorizeResponseError('%s: %s' % (error.code, error.text))
        return response

    def create_subscription(self, credit_card, amount, start,
            days=None, months=None, occurrences=None, trial_amount=None,
            trial_occurrences=None):
        """
        Creates a recurring subscription payment on the CreditCard provided.

        ``credit_card``
            The CreditCard instance to create the subscription for.
            Subscriptions require that you provide a first and last name with
            the credit card.

        ``amount``
            The amount to charge every occurrence, either as an int, float,
            or Decimal.

        ``start``
            The date to start the subscription, as a date object.

        ``days``
            Provide either the days or the months argument to indicate the
            interval at which the subscription should recur.

        ``months``
            Provide either the days or the months argument to indicate the
            interval at which the subscription should recur.

        ``occurrences``
            If provided, this is the number of times to charge the credit card
            before ending. If not provided, will last until canceled.

        ``trial_amount``
            If you want to have a trial period at a lower amount for this
            subscription, provide the amount. (Either both trial arguments
            should be provided, or neither.)

        ``trial_occurrences``
            If you want to have a trial period at a lower amount for this
            subscription, provide the number of occurences the trial period
            should last for. (Either both trial arguments should be provided,
            or neither.)
        """
        subscription = self.client.factory.create('ARBSubscriptionType')

        # Add the basic amount and payment fields
        amount = Decimal(str(amount)).quantize(Decimal('0.01'))
        subscription.amount = str(amount)
        payment_type = self.client.factory.create('PaymentType')
        credit_card_type = self.client.factory.create('CreditCardType')
        credit_card_type.cardNumber = credit_card.card_number
        credit_card_type.expirationDate = '{0}-{1:0>2}'.format(
            credit_card.exp_year, credit_card.exp_month)
        credit_card_type.cardCode = credit_card.cvv
        payment_type.creditCard = credit_card_type
        subscription.payment = payment_type
        if not (credit_card.first_name and credit_card.last_name):
            raise AuthorizeInvalidError('Subscriptions require first name '
                'and last name to be provided with the credit card.')
        subscription.billTo.firstName = credit_card.first_name
        subscription.billTo.lastName = credit_card.last_name

        # Add the fields for the payment schedule
        if (days and months) or not (days or months):
            raise AuthorizeInvalidError('Please provide either the months or '
                'days argument to define the subscription interval.')
        if days:
            try:
                days = int(days)
                assert days >= 7 and days <= 365
            except (AssertionError, ValueError):
                raise AuthorizeInvalidError('The interval days must be an '
                    'integer value between 7 and 365.')
            subscription.paymentSchedule.interval.unit = \
                self.client.factory.create('ARBSubscriptionUnitEnum').days
            subscription.paymentSchedule.interval.length = days
        elif months:
            try:
                months = int(months)
                assert months >= 1 and months <= 12
            except (AssertionError, ValueError):
                raise AuthorizeInvalidError('The interval months must be an '
                    'integer value between 1 and 12.')
            subscription.paymentSchedule.interval.unit = \
                self.client.factory.create('ARBSubscriptionUnitEnum').months
            subscription.paymentSchedule.interval.length = months
        if start < date.today():
            raise AuthorizeInvalidError('The start date for the subscription '
                'may not be in the past.')
        subscription.paymentSchedule.startDate = start.strftime('%Y-%m-%d')
        if occurrences is None:
            occurrences = 9999  # That's what they say to do in the docs
        subscription.paymentSchedule.totalOccurrences = occurrences

        # If a trial period has been specified, add those fields
        if trial_amount and trial_occurrences:
            subscription.paymentSchedule.trialOccurrences = trial_occurrences
            trial_amount = Decimal(str(trial_amount))
            trial_amount = trial_amount.quantize(Decimal('0.01'))
            subscription.trialAmount = str(trial_amount)
        elif trial_amount or trial_occurrences:
            raise AuthorizeInvalidError('To indicate a trial period, you '
                'must provide both a trial amount and occurrences.')

        # Make the API call to create the subscription
        response = self._make_call('ARBCreateSubscription', subscription)
        return response.subscriptionId

    def update_subscription(self, subscription_id, amount=None, start=None,
            occurrences=None, trial_amount=None, trial_occurrences=None):
        """
        Updates an existing recurring subscription payment. All fields to
        update are optional, and only the provided fields will be udpated.
        Many of the fields have particular restrictions that must be followed,
        as noted below.

        ``subscription_id``
            The subscription ID returned from the original create_subscription
            call for the subscription you want to update.

        ``amount``
            The updated amount to charge every occurrence, either as an int,
            float, or Decimal.

        ``start``
            The updated date to start the subscription, as a date object. This
            may only be udpated if no successful payments have been completed.

        ``occurrences``
            This updates the number of times to charge the credit card before
            ending.

        ``trial_amount``
            Updates the amount charged during the trial period. This may only
            be updated if you have not begun charging at the regular price.

        ``trial_occurrences``
            Updates the number of occurrences for the trial period. This may
            only be updated if you have not begun charging at the regular
            price.
        """
        subscription = self.client.factory.create('ARBSubscriptionType')

        # Add the basic subscription updates
        if amount:
            amount = Decimal(str(amount)).quantize(Decimal('0.01'))
            subscription.amount = str(amount)
        if start and start < date.today():
            raise AuthorizeInvalidError('The start date for the subscription '
                'may not be in the past.')
        if start:
            subscription.paymentSchedule.startDate = start.strftime('%Y-%m-%d')
        if occurrences:
            subscription.paymentSchedule.totalOccurrences = occurrences
        if trial_amount:
            trial_amount = Decimal(str(trial_amount))
            trial_amount = trial_amount.quantize(Decimal('0.01'))
            subscription.trialAmount = str(trial_amount)
        if trial_occurrences:
            subscription.paymentSchedule.trialOccurrences = trial_occurrences

        # Make the API call to update the subscription
        self._make_call('ARBUpdateSubscription', subscription_id,
            subscription)

    def delete_subscription(self, subscription_id):
        """
        Deletes an existing recurring subscription payment.

        ``subscription_id``
            The subscription ID returned from the original create_subscription
            call for the subscription you want to delete.
        """
        self._make_call('ARBCancelSubscription', subscription_id)



from decimal import Decimal
from six import text_type
from six.moves.urllib.parse import urlencode
from datetime import datetime
from ssl import SSLError

from suds import WebFault
from suds.client import Client
from authorize.data import Address, CreditCard

from authorize.apis.transaction import parse_response
from authorize.exceptions import AuthorizeConnectionError, \
    AuthorizeError, AuthorizeResponseError, AuthorizeInvalidError

PROD_URL = 'https://api.authorize.net/soap/v1/Service.asmx?WSDL'
TEST_URL = 'https://apitest.authorize.net/soap/v1/Service.asmx?WSDL'


class CustomerAPI(object):
    def __init__(self, login_id, transaction_key, debug=True, test=False):
        self.url = TEST_URL if debug else PROD_URL
        self.login_id = login_id
        self.transaction_key = transaction_key
        self.transaction_options = urlencode({
            'x_version': '3.1',
            'x_test_request': 'Y' if test else 'F',
            'x_delim_data': 'TRUE',
            'x_delim_char': ';',
        })

    @property
    def client(self):
        # Lazy instantiation of SOAP client, which hits the WSDL url
        if not hasattr(self, '_client'):
            self._client = Client(self.url)
        return self._client

    @property
    def client_auth(self):
        if not hasattr(self, '_client_auth'):
            self._client_auth = self.client.factory.create(
                'MerchantAuthenticationType')
            self._client_auth.name = self.login_id
            self._client_auth.transactionKey = self.transaction_key
        return self._client_auth

    def _make_call(self, service, *args):
        # Provides standard API call error handling
        method = getattr(self.client.service, service)
        try:
            response = method(self.client_auth, *args)
        except (WebFault, SSLError) as e:
            raise AuthorizeConnectionError('Error contacting SOAP API.')
        if response.resultCode != 'Ok':
            error = response.messages[0][0]
            e = AuthorizeResponseError('%s: %s' % (error.code, error.text))
            e.full_response = {
                'response_code': error.code,
                'response_text': error.text,
            }
            raise e
        return response

    def create_saved_profile(self, internal_id, payments=None, email=None):
        """
        Creates a user profile to which you can attach saved payments.
        Requires an internal_id to uniquely identify this user. If a list of
        saved payments is provided, as generated by create_saved_payment,
        these will be automatically added to the user profile. Returns the
        user profile id.
        """
        profile = self.client.factory.create('CustomerProfileType')
        profile.merchantCustomerId = internal_id
        profile.email = email
        if payments:
            payment_array = self.client.factory.create(
                'ArrayOfCustomerPaymentProfileType')
            payment_array.CustomerPaymentProfileType = payments
            profile.paymentProfiles = payment_array
        response = self._make_call('CreateCustomerProfile', profile, 'none')
        profile_id = response.customerProfileId
        payment_ids = None
        if payments:
            payment_ids = response.customerPaymentProfileIdList[0]
        return profile_id, payment_ids

    @staticmethod
    def _address_to_profile(address, payment_profile):
        if address and address.street:
            payment_profile.billTo.address = address.street
        if address and address.city:
            payment_profile.billTo.city = address.city
        if address and address.state:
            payment_profile.billTo.state = address.state
        if address and address.zip_code:
            payment_profile.billTo.zip = address.zip_code
        if address and address.country:
            payment_profile.billTo.country = address.country
        return payment_profile

    def create_saved_payment(self, credit_card, address=None, profile_id=None):
        """
        Creates a payment profile. If profile_id is provided, this payment
        profile will be created in Authorize.net attached to that profile.
        If it is not provided, the payment profile will be returned and can
        be provided in a list to the create_profile call.
        """
        # Create the basic payment profile with credit card details
        payment_profile = self.client.factory.create(
            'CustomerPaymentProfileType')
        customer_type_enum = self.client.factory.create('CustomerTypeEnum')
        payment_profile.customerType = customer_type_enum.individual
        payment_type = self.client.factory.create('PaymentType')
        credit_card_type = self.client.factory.create('CreditCardType')
        credit_card_type.cardNumber = credit_card.card_number
        credit_card_type.expirationDate = '{0.exp_year}-{0.exp_month:0>2}' \
            .format(credit_card)
        credit_card_type.cardCode = credit_card.cvv
        payment_type.creditCard = credit_card_type
        payment_profile.payment = payment_type

        # Customer billing name and address are optional fields
        if credit_card.first_name:
            payment_profile.billTo.firstName = credit_card.first_name
        if credit_card.last_name:
            payment_profile.billTo.lastName = credit_card.last_name
        payment_profile = self._address_to_profile(address, payment_profile)

        # If a profile id is provided, create saved payment on that profile
        # Otherwise, return an object for a later call to create_saved_profile
        if profile_id:
            response = self._make_call('CreateCustomerPaymentProfile',
                profile_id, payment_profile, 'none')
            return response.customerPaymentProfileId
        else:
            return payment_profile

    def retrieve_saved_payment(self, profile_id, payment_id):
        payment_id = int(payment_id)
        profile = self._make_call(
            'GetCustomerProfile', profile_id).profile
        payment_info = {}
        email = None
        if hasattr(profile, 'email'):
            email = text_type(profile.email)
        payment_info['email'] = email
        saved_payment = None
        for payment in profile.paymentProfiles[0]:
            if payment.customerPaymentProfileId == payment_id:
                saved_payment = payment
                break
        if not saved_payment:
            raise AuthorizeError("Payment ID does not exist for this profile.")
        payment_info['number'] = text_type(
            saved_payment.payment.creditCard.cardNumber)
        data = saved_payment.billTo
        payment_info['first_name'] = text_type(getattr(data, 'firstName', ''))
        payment_info['last_name'] = text_type(getattr(data, 'lastName', ''))
        kwargs = {
            'street': getattr(data, 'address', None),
            'city': getattr(data, 'city', None),
            'state': getattr(data, 'state', None),
            'zip_code': getattr(data, 'zip', None),
            'country': getattr(data, 'country', None)}
        kwargs = dict(
            [(key, text_type(value)) for key, value in kwargs.items() if value])
        payment_info['address'] = Address(**kwargs)
        return payment_info

    def update_saved_payment(self, profile_id, payment_id, **kwargs):
        payment_profile = self.client.factory.create(
            'CustomerPaymentProfileExType')
        customer_type_enum = self.client.factory.create('CustomerTypeEnum')
        payment_profile.customerType = customer_type_enum.individual
        payment_simple_type = self.client.factory.create('PaymentType')
        card_simple_type = self.client.factory.create('CreditCardSimpleType')
        number = kwargs['number']
        # Authorize.net uses this constant to indicate that we want to keep
        # the existing expiration date.
        date = 'XXXX'
        card_simple_type.cardNumber = number
        if kwargs['exp_month'] and kwargs['exp_year']:
            exp = CreditCard.exp_time(kwargs['exp_month'], kwargs['exp_year'])
            if exp <= datetime.now():
                raise AuthorizeInvalidError('This credit card has expired.')
            card_simple_type.expirationDate =\
                '{0}-{1:0>2}'.format(kwargs['exp_year'], kwargs['exp_month'])
        else:
            card_simple_type.expirationDate = date
        payment_simple_type.creditCard = card_simple_type
        payment_profile.payment = payment_simple_type
        payment_profile.payment.creditCard = card_simple_type
        payment_profile.customerPaymentProfileId = payment_id

        if kwargs['first_name']:
            payment_profile.billTo.firstName = kwargs['first_name']
        if kwargs['last_name']:
            payment_profile.billTo.lastName = kwargs['last_name']
        payment_profile = self._address_to_profile(
            kwargs['address'], payment_profile)

        self._make_call(
            'UpdateCustomerPaymentProfile', profile_id,
            payment_profile, 'none')

        if not kwargs['email']:
            return
        profile = self.client.factory.create('CustomerProfileExType')
        profile.email = kwargs['email']
        profile.customerProfileId = profile_id
        self._make_call('UpdateCustomerProfile', profile)

    def delete_saved_profile(self, profile_id):
        self._make_call('DeleteCustomerProfile', profile_id)

    def delete_saved_payment(self, profile_id, payment_id):
        self._make_call('DeleteCustomerPaymentProfile',
            profile_id, payment_id)

    def auth(self, profile_id, payment_id, amount, cvv=None):
        if cvv is not None:
            try:
                int(cvv)
            except ValueError:
                raise AuthorizeInvalidError("CVV Must be a number.")
        transaction = self.client.factory.create('ProfileTransactionType')
        auth = self.client.factory.create('ProfileTransAuthOnlyType')
        amount = Decimal(str(amount)).quantize(Decimal('0.01'))
        auth.amount = str(amount)
        auth.customerProfileId = profile_id
        auth.customerPaymentProfileId = payment_id
        auth.cardCode = cvv
        transaction.profileTransAuthOnly = auth
        response = self._make_call('CreateCustomerProfileTransaction',
            transaction, self.transaction_options)
        return parse_response(response.directResponse)

    def capture(self, profile_id, payment_id, amount, cvv=None):
        if cvv is not None:
            try:
                int(cvv)
            except ValueError:
                raise AuthorizeInvalidError("CVV Must be a number.")
        transaction = self.client.factory.create('ProfileTransactionType')
        capture = self.client.factory.create('ProfileTransAuthCaptureType')
        amount = Decimal(str(amount)).quantize(Decimal('0.01'))
        capture.amount = str(amount)
        capture.customerProfileId = profile_id
        capture.customerPaymentProfileId = payment_id
        capture.cardCode = cvv
        transaction.profileTransAuthCapture = capture
        response = self._make_call('CreateCustomerProfileTransaction',
            transaction, self.transaction_options)
        return parse_response(response.directResponse)

    def credit(self, profile_id, payment_id, amount):
        # Creates an "unlinked credit" (as opposed to refunding a previous transaction)
        transaction = self.client.factory.create('ProfileTransactionType')
        credit = self.client.factory.create('ProfileTransRefundType')
        amount = Decimal(str(amount)).quantize(Decimal('0.01'))
        credit.amount = str(amount)
        credit.customerProfileId = profile_id
        credit.customerPaymentProfileId = payment_id
        transaction.profileTransRefund = credit
        response = self._make_call('CreateCustomerProfileTransaction',
            transaction, self.transaction_options)
        return parse_response(response.directResponse)



from decimal import Decimal
from six import PY2, b, text_type
from six.moves.urllib.parse import urlencode
from six.moves.urllib.request import urlopen

from authorize.exceptions import AuthorizeConnectionError, \
    AuthorizeResponseError


PROD_URL = 'https://secure.authorize.net/gateway/transact.dll'
TEST_URL = 'https://test.authorize.net/gateway/transact.dll'
RESPONSE_FIELDS = {
    0: 'response_code',
    2: 'response_reason_code',
    3: 'response_reason_text',
    4: 'authorization_code',
    5: 'avs_response',
    6: 'transaction_id',
    9: 'amount',
    11: 'transaction_type',
    38: 'cvv_response',
}

DEFAULT_CHARSET = 'iso-8859-1'


def get_content_charset(resource):
    """Gets the charset encoding used in the given urlopen response."""
    if PY2:
        # Python 2 doesn't support get_content_charset, provides getparam
        # instead of get_param, and doesn't support the failobj option.
        charset = resource.headers.getparam('charset')
        return charset and charset.lower() or DEFAULT_CHARSET
    else:
        return resource.headers.get_content_charset(failobj=DEFAULT_CHARSET)


def parse_response(response):
    response = response.split(';')
    fields = {}
    for index, name in RESPONSE_FIELDS.items():
        fields[name] = response[index]
    return fields


def safe_unicode_to_str(string):
    try:
        return str(string)
    except UnicodeEncodeError:
        return string.encode('utf-8')


def convert_params_to_byte_str(params):
    converted_params = {}
    for key, value in params.items():
        if isinstance(key, text_type):
            key = safe_unicode_to_str(key)
        if isinstance(value, text_type):
            value = safe_unicode_to_str(value)
        converted_params[key] = value
    return converted_params


class TransactionAPI(object):
    def __init__(self, login_id, transaction_key, debug=True, test=False):
        self.url = TEST_URL if debug else PROD_URL
        self.base_params = {
            'x_login': login_id,
            'x_tran_key': transaction_key,
            'x_version': '3.1',
            'x_test_request': 'TRUE' if test else 'FALSE',
            'x_delim_data': 'TRUE',
            'x_delim_char': ';',
        }

    def _make_call(self, params):
        params = convert_params_to_byte_str(params)
        params = urlencode(params)
        try:
            resource = urlopen(self.url, data=b(params))
            response = resource.read().decode(
                get_content_charset(resource) or DEFAULT_CHARSET)
        except IOError as e:
            raise AuthorizeConnectionError(e)
        fields = parse_response(response)
        if fields['response_code'] != '1':
            e = AuthorizeResponseError(
                '{0} full_response={1!r}'.format(
                    fields['response_reason_text'], fields
                )
            )
            e.full_response = fields
            raise e
        return fields

    def _add_params(self, params, credit_card=None, address=None, email=None):
        if credit_card:
            params.update({
                'x_card_num': credit_card.card_number,
                'x_exp_date': credit_card.expiration.strftime('%m-%Y'),
                'x_card_code': credit_card.cvv,
                'x_first_name': credit_card.first_name,
                'x_last_name': credit_card.last_name,
            })

        if email:
            params['x_email'] = email

        if address:
            params.update({
                'x_address': address.street,
                'x_city': address.city,
                'x_state': address.state,
                'x_zip': address.zip_code,
                'x_country': address.country,
            })
        for key, value in list(params.items()):
            if value is None:
                del params[key]
        return params

    def auth(self, amount, credit_card, address=None, email=None):
        amount = Decimal(str(amount)).quantize(Decimal('0.01'))
        params = self.base_params.copy()
        params = self._add_params(params, credit_card, address, email)
        params['x_type'] = 'AUTH_ONLY'
        params['x_amount'] = str(amount)
        return self._make_call(params)

    def capture(self, amount, credit_card, address=None, email=None):
        amount = Decimal(str(amount)).quantize(Decimal('0.01'))
        params = self.base_params.copy()
        params = self._add_params(params, credit_card, address, email)
        params['x_type'] = 'AUTH_CAPTURE'
        params['x_amount'] = str(amount)
        return self._make_call(params)

    def settle(self, transaction_id, amount=None):
        # Amount is not required -- if provided, settles for a lower amount
        # than the original auth; if not, settles the full amount authed.
        params = self.base_params.copy()
        params['x_type'] = 'PRIOR_AUTH_CAPTURE'
        params['x_trans_id'] = transaction_id
        if amount:
            amount = Decimal(str(amount)).quantize(Decimal('0.01'))
            params['x_amount'] = str(amount)
        return self._make_call(params)

    def credit(self, card_num, transaction_id, amount):
        # Authorize.net can do unlinked credits (not tied to a previous
        # transaction) but we do not (at least for now).
        # Provide the last four digits for the card number, as well as the
        # transaction id and the amount to credit back.
        # The following restrictions apply:
        # - The transaction id must reference an existing, settled charge.
        #   (Note that in production, settlement happens once daily.)
        # - The amount of the credit (and the sum of all credits against this
        #   original transaction) must be less than or equal to the original
        #   charge amount.
        # - The credit must be submitted within 120 days of the original
        #   transaction being settled.
        params = self.base_params.copy()
        params['x_type'] = 'CREDIT'
        params['x_trans_id'] = transaction_id
        params['x_card_num'] = str(card_num)
        amount = Decimal(str(amount)).quantize(Decimal('0.01'))
        params['x_amount'] = str(amount)
        return self._make_call(params)

    def void(self, transaction_id):
        params = self.base_params.copy()
        params['x_type'] = 'VOID'
        params['x_trans_id'] = transaction_id
        return self._make_call(params)



"""
Authorize Sauce
===============

The secret sauce for accessing the Authorize.net API. The Authorize APIs for
transactions, recurring payments, and saved payments are all different and
awkward to use directly. Instead, you can use Authorize Sauce, which unifies
all three Authorize.net APIs into one coherent Pythonic interface. Charge
credit cards, easily!

::

  >>> # Init the authorize client and a credit card
  >>> from authorize import AuthorizeClient, CreditCard
  >>> authorize = AuthorizeClient('285tUPuS', '58JKJ4T95uee75wd')
  >>> cc = CreditCard('4111111111111111', '2018', '01', '911', 'Joe', 'Blow')
  >>> card = client.card(cc)

  >>> # Charge a card
  >>> card.capture(100)
  <AuthorizeTransaction 2171829470>

  >>> # Save the card on Authorize servers for later
  >>> saved_card = card.save()
  >>> saved_card.uid
  '7713982|6743206'

  >>> # Use a saved card to auth a transaction, and settle later
  >>> saved_card = client.saved_card('7713982|6743206')
  >>> transaction = saved_card.auth(200)
  >>> transaction.settle()

Saucy Features
--------------

* Charge a credit card
* Authorize a credit card charge, and settle it or release it later
* Credit or refund to a card
* Save a credit card securely on Authorize.net's servers
* Use saved cards to charge, auth and credit
* Create recurring charges, with billing cycles, trial periods, etc.

For the full documentation, please visit us at `Read the Docs`_. Thanks to
Chewse_ for supporting the development and open-sourcing of this library.
Authorize Sauce is released under the `MIT License`_.

.. _Read the Docs: http://authorize-sauce.readthedocs.org/
.. _Chewse: https://www.chewse.com/
.. _MIT License: http://www.opensource.org/licenses/mit-license
"""

import os
from setuptools import setup


# Hard links don't work inside VirtualBox shared folders. In order to allow
# setup.py sdist to work in such an environment, this quick and dirty hack is
# used. See http://stackoverflow.com/a/22147112.
if os.path.abspath(__file__).split(os.path.sep)[1] == 'vagrant':
    del os.link

setup(
    name='AuthorizeSauce',
    version='0.5.0',
    author='Jeff Schenck',
    author_email='jmschenck@gmail.com',
    url='http://authorize-sauce.readthedocs.org/',
    download_url='https://github.com/jeffschenck/authorizesauce',
    description='An awesome-sauce Python library for accessing the Authorize.net API. Sweet!',
    long_description=__doc__,
    license='MIT',
    install_requires=[
        'suds-jurko>=0.6',
        'six>=1.9.0',
    ],
    packages=[
        'authorize',
        'authorize.apis',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'License :: OSI Approved :: MIT License',
        'Topic :: Office/Business :: Financial',
        'Topic :: Internet :: WWW/HTTP',
    ],
)
